"use strict";

const Promise = require("bluebird");
const syncpipe = require("syncpipe");
const util = require("util");
const Result = require("@joepie91/result");

const ID = Symbol("dlayer-source object ID");
const AllowErrors = Symbol("dlayer-source allow-errors marker");

// TODO: Make more readable
// TODO: Refactor allowErrors logic so that it's actually part of the internal $getProperty implementation in dlayer itself, and this abstraction uses that tool?

module.exports = {
	withSources: function withSources(schemaObject) {
		let { $sources, ... rest } = schemaObject;
	
		let generatedProperties = syncpipe($sources ?? {}, [
			(_) => Object.entries(_),
			(_) => _.flatMap(([ source, properties ]) => {
				return Object.entries(properties).map(([ property, selector ]) => {
					let getter = function (_args, context) {
						return Promise.try(() => {
							if (properties[ID] != null) {
								let dataSource = context.sources[source];

								if (dataSource != null) {
									// console.log(`Calling source '${source}' with ID ${util.inspect(properties[ID])}`);
									return Result.wrapAsync(() => dataSource.load(properties[ID]));
								} else {
									throw new Error(`Attempted to read from source '${source}', but no such source is registered`);
								}
							} else {
								// FIXME: Better error message
								throw new Error(`Must specify a dlayer-source ID`);
							}
						}).then((result) => {
							// console.log(`Result [${source}|${util.inspect(properties[ID])}] ${util.inspect(result)}`);

							// The AllowErrors option is set when a source definition has its own way to deal with (allowable) errors. Instead of simply propagating the error for all affected attributes, it calls the attribute handlers with the Result (or returns `undefined` if only a property is specified).
							// TODO: How to deal with null results? Allow them or not? Make it an option?
							if (result.isOK && result.value() == null) {
								// TODO: Change implementation to allow `Result.ok(null|undefined)` but not `null|undefined` directly?
								throw new Error(`Null-ish result returned for ID ${util.inspect(properties[ID])} from source '${source}'; this is not allowed, and there is probably a bug in your code. Please file a ticket if you have a good usecase for null-ish results!`);
							} else if (properties[AllowErrors] === true && typeof selector !== "string") {
								// Custom selectors always receive the Result as-is
								return selector(result);
							} else if (result.isError) {
								if (properties[AllowErrors] === true) {
									return undefined;
								} else {
									// This is equivalent to a `throw`, and so we just propagate it
									return result;
								}
							} else {
								// This is to support property name shorthand used in place of a selector function
								if (typeof selector === "string") {
									return result.value()[selector];
								} else {
									return selector(result.value());
								}
							}
						});
					};

					return [ property, getter ];
				});
			}),
			(_) => Object.fromEntries(_)
		]);
	
		// NOTE: We always specify the generated properties first, so that properties can be overridden by explicit known values to bypass the source lookup, if needed by the implementation
		return {
			... generatedProperties,
			... rest
		};
	},
	ID: ID,
	AllowErrors: AllowErrors
};



